# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import sys
import json

from ipykernel.jsonutil import json_clean, encode_images

from IPython.core.interactiveshell import (
    InteractiveShell, InteractiveShellABC
)

from IPython.core.displaypub import DisplayPublisher
from traitlets import Type

class ProxyShellCaptureOutput(object):
    def formatMessage(self, message):
        try:
            jsMsg = eval(message)
            return json.dumps(jsMsg)
        except Exception as e:
            return """{{"text/plain":"{0}"}}""".format(message)

    def write(self, message):
        try:
            self.sys_stdout.write( message )
        except Exception as e:
            self.sys_stdout.write("got error" + str(e))
    
    def __enter__(self):       
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr
        
        sys.stdout = self
        sys.stderr = self
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value is not None:
            print("exc_type", exc_type)
            print("exc_value", str(exc_value))
            print("traceback", str(traceback))
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr

class ProxyDisplayPublisher(DisplayPublisher):
    def publish(self, data, metadata=None, source=None):
        if not isinstance(data, dict):
            raise TypeError('data must be a dict, got: %r' % data)
        if metadata is not None and not isinstance(metadata, dict):
            raise TypeError('metadata must be a dict, got: %r' % data)
        content = {}
        content['data'] = encode_images(data)
        content['metadata'] = metadata
        print(json.dumps(json_clean(content)))

class ProxyInteractiveShell(InteractiveShell):
    display_pub_class = Type(ProxyDisplayPublisher)
    def enable_gui(gui, kernel=None):
        pass

InteractiveShellABC.register(ProxyInteractiveShell)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from .seabornBaseDisplay import SeabornBaseDisplay
from pixiedust.display.chart.colorManager import Colors
from pixiedust.utils import Logger
import numpy as np
import pandas as pd
import seaborn as sns
import math

@PixiedustRenderer(id="histogram")
@Logger()
class sbHistogramDisplay(SeabornBaseDisplay):
    def supportsAggregation(self, handlerId):
        return False

    # TODO: add support for keys
    def supportsKeyFields(self, handlerId):
        return False
  
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    # no keys by default
    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def canRenderChart(self):
        return (True, None)

    def getNumFigures(self):
        return len(self.getValueFields())

    def acceptOption(self, optionName):
        if optionName == 'histoChartType':
            return False
        return True

    def matplotlibRender(self, fig, ax):
        rug=self.options.get("rug","false") == "true"
        kde=self.options.get("kde","true") == "true"
        defaultbin = math.sqrt(len(self.getWorkingPandasDataFrame().index))
        binsize = int(self.options.get('binsize', defaultbin))

        def plot(ax, valueField=None, color=None):
            data = self.getWorkingPandasDataFrame()[valueField]
            sns.distplot( data, ax=ax, bins=binsize, rug=rug, kde=kde,
                kde_kws={"label":"{0} KDE Estim".format(valueField)}, hist_kws={"label":"{0} Freq".format(valueField)},
                label=valueField,color = color
            )

        if len(self.getValueFields()) > 1:
            for j, valueField in enumerate(self.getValueFields()):
                plot(self.getAxItem(ax, j), valueField, Colors.colormap(1.*j/2))
        else:
            plot(ax, self.getValueFields()[0])

    @commonChartOptions
    def getChartOptions(self):
        return [{
            'name': 'rug',
            'description': 'Rugplot',
            'metadata': {
                'type': "checkbox",
                'default': "false"
            }
        },{ 
            'name': 'kde',
            'description': 'Kernel Density',
            'metadata': {
                'type': "checkbox",
                'default': "false"
            }
        }]


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.colorManager import Colors
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
from bokeh.plotting import figure
import pandas as pd
import numpy as np
import sys

try:
    from bkcharts import Line
except ImportError:
    from bokeh.charts import Line

@PixiedustRenderer(id="lineChart")
@Logger()
class LineChartRenderer(BokehBaseDisplay):
    def isSubplot(self):
        return self.options.get("lineChartType", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def safeList(self, index):
        if np.issubdtype(index.dtype, np.datetime64):
            dateFormat = self.options.get("dateFormat", None)
            def convert(x):
                return str(x).replace(':','-') if dateFormat is None else x.strftime(dateFormat)
            return [convert(pd.to_datetime(x)) for x in index.values]
        return index.values.tolist()

    def createBokehChart(self):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        clusterby = self.options.get("clusterby")
        subplots = self.isSubplot()
        workingPDF = self.getWorkingPandasDataFrame().copy()

        for index, row in workingPDF.iterrows():
            for k in keyFields:
                if isinstance(row[k], str if sys.version >= '3' else basestring):
                    row[k] = row[k].replace(':', '.')
            workingPDF.loc[index] = row

        charts=[]
        if clusterby is not None and (subplots or len(valueFields)<=1):
            subplots = subplots if len(valueFields)==1 or subplots else False
            for j, valueField in enumerate(valueFields):
                pivot = workingPDF.pivot(
                    index=keyFields[0], columns=clusterby, values=valueField
                )

                if not subplots:
                    fig = figure(x_range=self.safeList(pivot.index) if not np.issubdtype(pivot.index.dtype, np.number) else None)
                    charts.append(fig)
                for i,col in enumerate(pivot.columns[:10]): #max 10
                    if subplots:
                        charts.append( 
                            Line(
                                pivot[col].values, color=Colors.hexRGB( 1.*i/2 ), ylabel=valueField, xlabel=keyFields[0], legend=False, 
                                title="{0} = {1}".format(clusterby, pivot.columns[i])
                            )
                        )
                    else:
                        xValues = pivot.index.values.tolist()
                        if not np.issubdtype(pivot.index.dtype, np.number):
                            xValues = range(1, len(xValues)+1)
                        fig.line(x = xValues, y = pivot[col].values, color = Colors.hexRGB( 1.*i/2 ), legend=str(col) if self.showLegend() else None)
        else:
            if subplots:
                for i,valueField in enumerate(valueFields):
                    charts.append(Line(workingPDF, x = keyFields[0], y=valueField, color = Colors.hexRGB( 1.*i/2 ), legend=self.showLegend(), plot_width=int(800/len(valueFields))))
            else:
                charts.append(Line(workingPDF, x = keyFields[0], y=valueFields, color=valueFields, legend=self.showLegend()))

            if clusterby is not None:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")
        return charts


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import *
from pixiedust.utils.dataFrameAdapter import *
import pixiedust.utils.dataFrameMisc as dataFrameMisc
    
class TableDisplay(Display):
    def doRender(self, handlerId):
        entity=self.entity       
        if dataFrameMisc.fqName(entity) == "graphframes.graphframe.GraphFrame":
            if handlerId == "edges":
                entity=entity.edges
            else:
                entity=entity.vertices
        if dataFrameMisc.isPySparkDataFrame(entity) or dataFrameMisc.isPandasDataFrame(entity):
            hcmap = {}
            if "hideColumns" in self.options:
                hcarr = self.options.get("showColumns").split(",")
                for s in hcarr:
                    hcmap[s] = 1
            self._addHTMLTemplate('dataframeTable.html', entity=PandasDataFrameAdapter(entity), table_noschema=self.options.get("table_noschema", "false"), table_hidecols=hcmap)
            return
  
        self._addHTML("""
            <b>Unable to display object {}</b>
        """.format(entity)
        )



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.dataFrameAdapter import PandasDataFrameAdapter
import pixiedust.utils.dataFrameMisc as dataFrameMisc
from datetime import datetime
import pandas as pd
import numpy as np
from pixiedust.utils import Logger
from six import iteritems

@Logger()
class PandasDataFrameDataHandler(object):
    def __init__(self, options, entity):
        self.options = options
        self.entity = entity

    def getFieldNames(self, expandNested=False):
        return dataFrameMisc.getFieldNames(PandasDataFrameAdapter(self.entity), expandNested)

    def isNumericField(self, fieldName):
        for y in self.entity.columns:
            if y == fieldName:
                return self.entity[y].dtype == np.float64 or self.entity[y].dtype == np.int64
        raise ValueError("Column {} does not existing in the dataframe".format(fieldName))

    def isNumericType(self, field):
        return dataFrameMisc.isNumericType(field.dataType)

    def isStringField(self, fieldName):
        return dataFrameMisc.isStringField(PandasDataFrameAdapter(self.entity), fieldName)

    def isStringType(self, field):
        return dataFrameMisc.isStringType(field.dataType)

    def isDateField(self, fieldName):
        return dataFrameMisc.isDateField(PandasDataFrameAdapter(self.entity), fieldName)

    def isDateType(self, field):
        return dataFrameMisc.isDateType(field.dataType)

    @property
    def schema(self):
        return PandasDataFrameAdapter(self.entity).schema

    """
        Return a cleaned up Pandas Dataframe that will be used as working input to the chart
    """
    def getWorkingPandasDataFrame(self, xFields, yFields, extraFields=[], aggregation=None, maxRows = 100):
        if xFields is None or len(xFields)==0:
            #swap the yFields with xFields
            xFields = yFields
            yFields = []
            aggregation = None

        extraFields = [a for a in extraFields if a not in xFields and a not in yFields]
        workingDF = self.entity[xFields + extraFields + yFields]

        if aggregation and len(yFields)>0:
            aggMapper = {"SUM":"sum", "AVG": "mean", "MIN": "min", "MAX": "max"}
            aggFn = aggMapper.get(aggregation, "count")
            workingDF = workingDF.groupby(extraFields + xFields).agg(aggFn).reset_index()

        workingDF = workingDF.dropna()
        count = len(workingDF.index)
        if count > maxRows:
            workingDF = workingDF.sample(frac=(float(maxRows) / float(count)),replace=False)

        #check if the caller want to preserve some columns
        preserveCols = self.options.get("preserveCols", None)
        if preserveCols is not None:
            preserveCols = [a for a in preserveCols.split(",") if a not in xFields and a not in yFields]
            cols = { key:[] for key in preserveCols}
            for i in workingDF.index:
                cond = None
                for j, key in enumerate(extraFields + xFields):
                    thisKey = self.entity.loc[i][key]
                    cond = (self.entity[key] == thisKey) if j is 0 else (cond & (self.entity[key] == thisKey))
                p = self.entity[ cond ]
                for key in preserveCols:
                    cols[key].append( p[key].values.tolist()[0] )
            for key, value in iteritems(cols):
                workingDF.insert(len(workingDF.columns), key, value)

        #check if the user wants timeseries
        if len(xFields) == 1 and self.options.get("timeseries", 'false') == 'true':
            field = xFields[0]
            try:
                inputDateFormat = self.options.get("inputDateFormat", None)
                if inputDateFormat is not None:
                    workingDF[field] = workingDF[field].apply(lambda x: datetime.strptime(str(x), inputDateFormat))
                else:
                    workingDF[field] = pd.to_datetime(workingDF[field])
            except:
                self.exception("Unable to convert field {} to datetime".format(field))
        
        #sort by xFields
        workingDF.sort_values(extraFields + xFields, inplace=True)
        
        return workingDF


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import Display
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
import yaml
import pixiedust
import json

myLogger = pixiedust.getLogger(__name__)

class GraphDisplay(Display):
    def doRender(self,handlerId):
        g = self.entity
        width = int(self.getPreferredOutputWidth() - 10 )
        height = int(self.getPreferredOutputHeight() - 10  )

        if handlerId == "graphMap":
            graphNodesJson="{"

            for r in g.vertices.rdd.map(lambda row: """"{0}":{{"id":"{0}","name":"{1}","latitude":{2},"longitude":{3}}}"""
                .format(row.id, row.name.encode("ascii","ignore").decode("ascii"),0.0 if row.latitude is None else row.latitude,0.0 if row.longitude is None else row.longitude)).collect():
                graphNodesJson+=("," if len(graphNodesJson)>1 else "") + str(r)

            graphNodesJson+="}"

            graphLinksJson = str(g.edges.select("src","dst").groupBy("src","dst").agg(F.count("src").alias("count")).toJSON().map(lambda j: yaml.safe_load(j)).collect())
            
            myLogger.debug("graphMap - nodes: {0}".format(graphNodesJson))
            myLogger.debug("graphMap - links: {0}".format(graphLinksJson))

            self._addScriptElement("https://d3js.org/d3.v3.js", checkJSVar="d3",
                callback=self.renderTemplate("graphMap.js", graphNodesJson=graphNodesJson, graphLinksJson=graphLinksJson, preferredWidth=width, preferredHeight=height))

            self._addHTMLTemplate("graphMap.html")

        elif handlerId == "graphTree":
            def expand(values, visited, level):
                results=[]
                if values is not None and level < maxDepth:
                    for v in values:
                        if v not in visited and len(results)<maxChildren:
                            visited[v]=True
                            results.append({ "name": str(v), "children": {}})
                    for item in results:
                        nextVisited = {}
                        nextVisited.update(visited)
                        item["children"]=expand(dic.get(item["name"]), nextVisited, level+1)
                return results
            
            ar = g.edges.select("src","dst").rdd.map(lambda row: (row[0],[row[1]]))\
                .reduceByKey(lambda d1,d2: d1+d2).map(lambda row: (row[0], list(set(row[1]))))\
                .collect()

            dic = {item[0] : item[1] for item in ar}
            maxDepth = self.options.get("maxDepth", 5)
            maxChildren = self.options.get("maxChildren", 10)
            root = self.options.get("root")
            rootNode = ar[0]

            if root:
                def findRoot(ar):
                    for a in ar:
                        if a[0]==root:
                            return a
                rootNode = findRoot(ar)
            
            if not rootNode:
                self._addHTML("<p>Root node not found!</p>")
            else:
                res = { "name": str(rootNode[0]), "children":expand(dic[ar[0][0]], {rootNode[0]:True}, 1)}
                tree = json.dumps(res)

                myLogger.debug("graphTree - tree: {0}".format(res))

                #if user specified root, then only send back the json tree
                if root:
                    self.addProfilingTime = False
                    print(tree)
                else:
                    nodes = g.vertices.select('id').orderBy('id').rdd.map(lambda r: r[0]).collect()
                    self._addScriptElement("https://d3js.org/d3.v3.js", checkJSVar="d3", 
                        callback=self.renderTemplate("graphTree.js", root=str(rootNode[0]), tree=tree, preferredWidth=width, preferredHeight=height))
                    self._addHTMLTemplate("graph.html", root=str(rootNode[0]), nodes=nodes, maxDepth=maxDepth, maxChildren=maxChildren, handlerId=handlerId)
            
        else:
            # force-directed graph
            maxEdges = self.options.get("maxEdges", 100)
            cols = [g.edges.columns[i] for i in range(len(g.edges.columns)) if g.edges.columns[i] not in ['src', 'dst']]
            edges = g.edges.toPandas()[:maxEdges].to_json(orient='records')
            graph = json.dumps(edges)
            isupdate = self.options.get("isupdate")

            cols.sort()
            colorBy = self.options.get("colorBy", cols[0] if len(cols) > 0 else "")

            myLogger.debug("graphDirected - edges: {0}".format(edges))

            #if user specified update, then only send back the json graph
            if isupdate:
                self.addProfilingTime = False
                print(graph)
            else:
                self._addScriptElement("https://d3js.org/d3.v3.js", checkJSVar="d3",
                    callback=self.renderTemplate("graphDirected.js", graph=graph, preferredWidth=width, preferredHeight=height, colorBy=colorBy))
                self._addHTMLTemplate("graph.html", maxEdges=maxEdges, handlerId=handlerId, cols=cols, colorBy=colorBy)



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from .cfBrowser import CFBrowser
from pixiedust.display.app import *
from pixiedust.services.serviceManager import *
from pixiedust.utils import Logger

@Logger()
class ConnectionWidget(CFBrowser):
    def getConnections(self):
        return getConnections("cloudant")

    def selectBluemixCredentials(self, service_name, credentials_str):
        credentials = json.loads(credentials_str)
        payload = {'name': service_name, 'credentials': credentials}
        addConnection('cloudant', payload)
        self.selectedConnection = payload['name']
        #return self.dataSourcesList()
        return """
<script>
    pixiedust.executeDisplay({{pd_controls}}, {
        'targetDivId': "dummy",
        'script': "import json\\nprint(json.dumps( self.getConnections()))",
        'onError': function(error){
            alert(error);
        },
        'onSuccess': function(results){
            var options = []
            $.each(JSON.parse(results), function(key, value){
                var selected = (value.name=="{{this.selectedConnection}}") ? 'selected="selected"' : "";
                options.push('<option ' + selected + ' value="'+ value.name +'">'+ value.name +'</option>');
            });
            $("#connection{{prefix}}").html(options.join(''));
        }
    });
</script>
        """
    
    @route(selectedConnection="*", editConnection="*")
    def _editConnection(self):
        jsOnLoad = """
var CodeMirror = require('codemirror/lib/codemirror');
function createEditor(content){
    $('#connTextArea{{prefix}}').html('<textarea id="connection-info" rows=13 cols=80 name="connection-info"/>');
    global.editor = CodeMirror.fromTextArea($("#connection-info")[0], {
        lineNumbers: true,
        matchBrackets: true,
        indentUnit: 2,
        autoIndent: true,
        mode: 'application/json'
    });
    global.editor.setSize("100%", 300);
    global.editor.setValue(content);
}
var connectionName = $("#connection{{prefix}}").val();
if ( connectionName ){
    function toJson(v){
        if (!v){
            return ""
        }
        v = JSON.parse(v);
        return JSON.stringify(v, null, '\t');
    }
    var script = "from pixiedust.services.serviceManager import getConnection\\n"+
        "import json\\n"+
        "print( json.dumps( getConnection(\\\"cloudant\\\",\\\"" + connectionName + "\\\", raw=False)))";
    debugger;
    pixiedust.executeDisplay({{pd_controls}}, {
        'targetDivId': "connTextArea{{prefix}}",
        'script': script,
        'onError': function(error){
            $("#connection-error").text(error);
        },
        'onSuccess': function(results){
            debugger;
            createEditor(toJson(results));
        }
    })
}else{
    createEditor('{\\n\\
        "name": "<<cloudant-connection name>>",\\n\\
        "credentials": {\\n\\
            "username": "<<userIdentifier>>-bluemix",\\n\\
            "password": "<<password>>",\\n\\
            "host": "<<hostIdentifier>>-bluemix.cloudant.com",\\n\\
            "port": 443,\\n\\
            "url": "https://<<userIdentifier>>-bluemix:<<password>>@<<hostIdentifier>>-bluemix.cloudant.com"\\n\\
        }\\n\\
    }');
}
        """
        jsOK = """
debugger;
try {
    var connInfo = JSON.parse(global.editor.getValue());
    var script = "from pixiedust.services.serviceManager import addConnection\\n"+
        "print(addConnection(\\"cloudant\\"," + global.editor.getValue().replace(/[\\n\\r]+/g,"") + "))";
    pixiedust.executeDisplay({{pd_controls}}, {
        'targetDivId': "dummy",
        'script': script,
        'onError': function(error){
            $("#connection-error").text(error);
        },
        'onSuccess': function(results){
            if (typeof editConnCallback != "undefined"){
                editConnCallback(results);
            }
            modal_obj.modal('hide')
        }
    });
    return false;
} catch(e) {
    console.log(e);
    $("#connection-error").text('System error: ' + e.message);
    return false;
}
        """
        body = """
<div class="well">
    Please enter your connection information as JSON in the text area below.
    For a new connection, you can for example copy and paste the service credentials from Bluemix.
</div>
<div id="connection-error" style="color:red"/>
<div id="connTextArea{{prefix}}"/>
        """
        return {"body":body, "jsOnLoad": jsOnLoad, "jsOK":jsOK}
    
    @route(selectedConnection="*", deleteConnection="true")
    def _deleteConnection(self):
        deleteConnection("cloudant", self.selectedConnection)
        self.deleteConnection = "false"
        return self.dataSourcesList()
        
    @route(selectedConnection="*", browseBMConnection="true")
    def _browseBMConnection(self):
        return self.startBrowsingBM()
    
    @route(selectedConnection="*", newConnection="true")
    def _newConnection(self):            
        body = """
<div class="well">
    New Connection
</div>
<div class="container" id="newConnectionContainer{{prefix}}">
    <div class="col-sm-2"/>
    <div class="col-sm-4" style="border: 1px solid lightblue;margin: 10px;border-radius: 25px;cursor:pointer;
        min-height: 150px;text-align: center;background-color:#e6eeff;display: flex;align-items: center;"
        pd_options="manualConnection=true">
        <h2>Enter the connection manually</h2>
    </div>
    <div class="col-sm-4" style="border: 1px solid lightblue;margin: 10px;border-radius: 25px;cursor:pointer;
        min-height: 150px;text-align: center;background-color:#e6eeff;;display: flex;align-items: center;"
        pd_options="browseBMConnection=true">
        <h2>Browse your services on Bluemix</h2>
    </div>
    <div class="col-sm-2"/>
</div>
"""
        return {"body":body, "dialogRoot": "newConnectionContainer{{prefix}}"}
    
    @route(widget="DataSourcesList")
    def dataSourcesList(self):
        num_connections = len(self.getConnections())
        select_conn_script = ' pd_script="self.selectedConnection ='
        if num_connections > 0:
            select_conn_script += "'$val(connection{{prefix}})'\""
        else:
            select_conn_script += '\'none\'"'
        output = """
<div>
    <div class="form-group">
      <label for="connection{{prefix}}" class="control-label col-sm-2">Select a cloudant connection:</label>
      <div class="col-sm-7">
        <select class="form-control" id="connection{{prefix}}">
          {%for conn in this.getConnections() %}
              {%set selected=(this.selectedConnection==conn.name)%}
              <option {%if selected%}selected="selected"{%endif%}  value="{{conn.name|escape}}">{{conn.name|escape}}</option>
          {%endfor%}
        </select>
      </div>
      <div class="col-sm-2 btn-toolbar" role="toolbar">
        <div class="btn-group" role="group\"""" + select_conn_script + """>
"""
        if num_connections > 0:
            output += """
            <button type="button" class="btn btn-default">Go</button>'
            <button type="button" class="btn btn-default" pixiedust pd_options="dialog=true;editConnection=true">
                <i class="fa fa-pencil-square-o"/>
            </button>"""
        output += """
            <button type="button" class="btn btn-default" pixiedust pd_options="dialog=true;newConnection=true">
                <i class="fa fa-plus"/>
            </button>"""
        if num_connections > 0:
            output += """
            <button type="button" class="btn btn-default" pixiedust>
                <pd_script type="preRun">
                    return confirm("Delete " + $("#connection{{prefix}}").val() + "?");
                </pd_script>
                <pd_script>self.deleteConnection="true"</pd_script>
                <i class="fa fa-trash"/>
            </button>
        </div>
      </div>
    </div>
</div>
"""
        return output


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.app import *
import requests

@PixieApp
class MapboxBase():
    def setLayers(self, layers):
        self.layers = layers

    def getStyleTypeFromGeoJSON(self, layerDef, geoJSON):
        styleType = geoJSON['features'][0]['geometry']['type']
        if styleType == "Point":
            userStyleType = layerDef.get("type", "circle")
            if userStyleType == "symbol":
                return ("symbol", 
                    self.mergeDef( layerDef.get("paint"), {}), 
                    self.mergeDef( layerDef.get("layout"), {})
                )
            else:
                return ("circle", self.mergeDef( layerDef.get("paint"), {
                    "circle-color": "rgba(255,0,0,0.5)", 
                    "circle-radius": 5, 
                    "circle-blur": 0
                }), self.mergeDef( layerDef.get("layout"), {}))
        elif styleType == "MultiPolygon":
            return ("fill", self.mergeDef( layerDef.get("paint"), {
                "fill-color": "rgb(125, 125, 0)",
                "fill-opacity": 0.25,
                "fill-outline-color": "rgba(0,0,255, 0.5)"
            }), self.mergeDef( layerDef.get("layout"), {}))
        return ("line", self.mergeDef(layerDef.get("paint"), {
                "line-color": "rgba(128,0,128,0.65)",
                "line-width": 16,
                "line-blur": 2,
                "line-opacity": 0.75
            }), self.mergeDef( layerDef.get("layout"), {}))

    def mergeDef(self, userDef, defaultDef):
        if userDef is not None:
            defaultDef.update(userDef)
        return defaultDef
        
    def createMapboxGeoJSON(self, order, layerDef, geoJSON):
        id = layerDef["name"]
        style = self.getStyleTypeFromGeoJSON(layerDef, geoJSON)
        return {
            "id": id, 
            "maptype":"mapbox",
            "order": order,
            "source":{
                "type": "geojson",
                "data": geoJSON
            },
            "type": style[0],
            "paint": style[1],
            "layout": style[2]
        }

    def toggleLayer(self, index):
        fieldName = "layer{}".format(index);
        if hasattr(self, fieldName) and getattr(self, fieldName) is not None:
            setattr(self, fieldName, None)
        else:
            setattr(self, fieldName, self.createMapboxGeoJSON(index+2, self.layers[index], self.loadGeoJSON( self.layers[index]["url"] ) ))        
        
    def loadGeoJSON(self, url):
        def filterFeature(f):
            for key,value in iteritems(f):
                if value is None or "'" in value:
                    return True
            return False
        payload = requests.get(url).json()
        payload['features'] = [f for f in payload['features'] if not filterFeature(f['properties'])]
        return payload


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from .seabornBaseDisplay import SeabornBaseDisplay
import pandas as pd
import numpy as np
import seaborn as sns

@PixiedustRenderer(id="scatterPlot")
class ScatterPlotDisplay(SeabornBaseDisplay):
	def supportsAggregation(self, handlerId):
		return False
    
	def canRenderChart(self):
		valueFields = self.getValueFields()
		if len(valueFields) != 1:
			return (False, "Can only specify one Value Field")

		#Verify that all key field are numericals
		for keyField in self.getKeyFields():
			if not self.dataHandler.isNumericField(keyField):
				return (False, "Column {0} is not numerical".format(keyField))
		
		return (True, None)

	def getPreferredDefaultValueFieldCount(self, handlerId):
		return 2

	def createFigure(self):
		valueFields = self.getValueFields()
		keyFields = self.getKeyFields()
		facetGrid = sns.jointplot(x=keyFields[0], y=valueFields[0], kind=self.options.get("kind","scatter"), data=self.getWorkingPandasDataFrame())
		return facetGrid.fig, facetGrid.fig.axes[0]

	def matplotlibRender(self, fig, ax):
		pass

	@commonChartOptions
	def getChartOptions(self):
		return [
			{ 'name': 'kind',
			  'metadata': {
					'type': "dropdown",
					'values': ["scatter", "reg", "resid", "kde", "hex"],
					'default': "scatter"
				}
			}
		]


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger

try:
    from bkcharts import Scatter
except ImportError:
    from bokeh.charts import Scatter

@PixiedustRenderer(id="scatterPlot")
@Logger()
class ScatterPlotRenderer(BokehBaseDisplay):

    def supportsAggregation(self, handlerId):
        return False

    @commonChartOptions
    def getChartOptions(self):
        return [
            { 'name': 'color',
              'metadata': {
                    'type': "dropdown",
                    'values': ["None"] + [f for f in self.getFieldNames() if f not in self.getKeyFields() and f not in self.getValueFields()],
                    'default': ""
                },
              'validate': lambda option:\
                    (option in self.getFieldNames() and option not in self.getKeyFields() and option not in self.getValueFields(),\
                    "color value is already used in keys or values for this chart")
            }
        ]

    def canRenderChart(self):
        valueFields = self.getValueFields()
        if len(valueFields) != 1:
            return (False, "Can only specify one Value Field")

        #Verify that all key field are numericals
        for keyField in self.getKeyFields():
            if not self.dataHandler.isNumericField(keyField):
                return (False, "Column {0} is not numerical".format(keyField))
        
        return (True, None)

    def getExtraFields(self):
        color = self.options.get("color")
        return [color] if color is not None else []

    def createBokehChart(self):        
        data = self.getWorkingPandasDataFrame()
        return Scatter(data, 
            x = self.getKeyFields()[0], y = self.getValueFields()[0],
            xlabel=self.getKeyFields()[0],ylabel=self.getValueFields()[0],legend=self.showLegend(), color=self.options.get("color"))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.display.chart.colorManager import Colors
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
from bokeh.layouts import row
import math

try:
    from bkcharts import Histogram, show
except ImportError:
    from bokeh.charts import Histogram, show

@PixiedustRenderer(id="histogram")
@Logger()
class HistogramRenderer(BokehBaseDisplay):
    
    def supportsAggregation(self, handlerId):
        return False

    def supportsKeyFields(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def acceptOption(self, optionName):
        if optionName == 'histoChartType':
            return False
        return True

    @commonChartOptions
    def getChartOptions(self):
        return [
            { 'name': 'color',
              'metadata': {
                    'type': "dropdown",
                    'values': ["None"] + self.getFieldNames(),
                    'default': ""
                }
            }
        ]

    def canRenderChart(self):
        return (True, None)

    def getExtraFields(self):
        color = self.options.get("color")
        return [color] if color is not None else []
    
    def createBokehChart(self):
        plotwidth = int(self.getPreferredOutputWidth()/2)
        plotheight = int(self.getPreferredOutputHeight() * 0.75)
        defaultbin = math.sqrt(len(self.getWorkingPandasDataFrame().index))
        binsize = int(self.options.get('binsize', defaultbin))
        valueFields = self.getValueFields()
        histograms = []
        if len(valueFields) != 1:
            for i,valueField in enumerate(valueFields):
                color = self.options.get("color")
                if color is None:
                    color = Colors.hexRGB( 1.*i/2 )
                histograms.append(Histogram(
                    self.getWorkingPandasDataFrame(), values=valueField, plot_width=plotwidth, plot_height=plotheight,bins=binsize, 
                    color=color, xgrid=True, ygrid=True, ylabel='Frequency')
                )
            return histograms
        else:
            return Histogram(self.getWorkingPandasDataFrame(), values=self.getValueFields()[0],
            bins=binsize, color=self.options.get("color"), 
            xgrid=True, ygrid=True, ylabel='Frequency')


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display.display import Display
from .serviceManager import *
import time
import requests
import json
from pixiedust.utils import Logger
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.environment import Environment

CLOUDANT_CONN_TYPE = "cloudant"
@Logger()
class StashCloudantHandler(Display):
    def tuplize(self, connections):
        return [tuple(connections[i:i+2]) for i in range(0, len(connections), 2)]

    def doRender(self, handlerId):
        if self.options.get("nostore_listConnections"):
            return self._addHTMLTemplate("listConnections.html", connections=self.tuplize(getConnections(CLOUDANT_CONN_TYPE)) )

        config = ShellAccess.sc._conf.getAll()
        checks = ["spark-cloudant", "cloudant-spark"]
        sparkJars = [t[1] for t in config if t[0]=="spark.jars"]
        cond1 = any("spark.jars" in s for s in config) and any( c in s for s in sparkJars for c in checks)
        cond2 = any(c in s for s in Environment.javaClassPath.split(":") for c in checks)
        if not cond1 and not cond2:
            self._addHTML("<p>Please setup one time manual steps to configure stashing to Cloudant database at <a href=\"https://ibm-cds-labs.github.io/pixiedust/install.html#stash-to-cloudant\"  target=\"_blank\">https://ibm-cds-labs.github.io/pixiedust/install.html#stash-to-cloudant</a></p>")
            self.debug("Please setup one time manual steps to configure stashing to Cloudant database at https://ibm-cds-labs.github.io/pixiedust/install.html#stash-to-cloudant")
            self.debug("SparkContext conf:")
            self.debug(config)
            return
                    
        entity=self.entity

        dbName = self.options.get("nostore_dbName", "dataframe-"+time.strftime('%Y%m%d-%H%M%S'))
        connectionName=self.options.get("nostore_connection")
        if connectionName is None:
            self._addHTMLTemplate("stashCloudant.html",dbName=dbName,connType=CLOUDANT_CONN_TYPE, connections=self.tuplize(getConnections(CLOUDANT_CONN_TYPE)))
        else:
            #first create the stash db
            connection = getConnection(CLOUDANT_CONN_TYPE, connectionName)
            if not connection:
                raise Exception("Unable to resolve connection {0}".format(connectionName))
            payload = json.loads( connection["PAYLOAD"])
            credentials=payload["credentials"]
            self.debug("write to cloudant host {0}".format(credentials["host"]))
            format = self.entity.write.format("com.cloudant.spark")\
                .option("cloudant.host", credentials["host"])\
                .option("cloudant.username",credentials["username"])\
                .option("cloudant.password",credentials["password"])\
                .option("createDBOnSave","true")
            
            if "protocol" in credentials:
                format.option("cloudant.protocol", credentials["protocol"])
            
            format.save(dbName)
            self._addHTML("""<div>Successfully stashed your data: <a target='_blank' href='{0}/{1}'>{1}</a></div>""".format(credentials["url"],dbName))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from ..baseChartDisplay import BaseChartDisplay
from six import with_metaclass
from abc import ABCMeta

import pixiedust

myLogger = pixiedust.getLogger(__name__)

@PixiedustRenderer(rendererId="mapbox")
class MapBoxBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
	pass

	def getChartOptions(self):
		return [ 
            {
                'name': 'legend',
                'description': 'Show legend',
                'metadata': {
                    'type': 'checkbox',
                    'default': "true"
                }
            },
			{
					'name': 'chartsize',
					'description': 'Map Size',
					'metadata': {
							'type': 'slider',
							'max': 100,
							'min': 50,
							'default': 90
					}
			},
			{ 'name': 'kind',
			  'metadata': {
					'type': 'dropdown',
					'values': ['simple','simple-cluster','choropleth','choropleth-cluster','densitymap'],
					'default': 'simple'
				}
			},
			{'name': 'basemap', 
			 'metadata': {
				 'type': 'dropdown', 
				 'values': ['light-v9', 'satellite-v9', 'dark-v9','outdoors-v9'], 
				 'default': 'light-v9'
			 }}
		]

    


from setuptools import setup, find_packages
setup(name='pixiedust',
	  version='1.0.7',
	  description='Productivity library for Spark Python Notebook',
	  url='https://github.com/ibm-cds-labs/pixiedust',
	  install_requires=['mpld3','lxml','geojson'],
	  author='David Taieb',
	  author_email='david_taieb@us.ibm.com',
	  license='Apache 2.0',
	  packages=find_packages(exclude=('tests', 'tests.*')),
	  include_package_data=True,
	  zip_safe=False,
	  entry_points={
		  'console_scripts': [
			  'jupyter-pixiedust = install.pixiedustapp:main'
		  ]		  
	  }
)



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.utils import Logger
from ..baseChartDisplay import BaseChartDisplay
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from six import with_metaclass
from abc import abstractmethod, ABCMeta
import numpy as np
import math
import matplotlib.ticker as ticker

try:
    import mpld3
    import mpld3.plugins as plugins
    from pixiedust.display.chart.plugins.chart import ChartPlugin
    from pixiedust.display.chart.plugins.dialog import DialogPlugin
    from pixiedust.display.chart.plugins.elementInfo import ElementInfoPlugin
    mpld3Available = True
except ImportError:
    mpld3Available = False

@PixiedustRenderer(rendererId="matplotlib")
@Logger()
class MatplotlibBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):

    def __init__(self, options, entity, dataHandler=None):
        super(MatplotlibBaseDisplay,self).__init__(options,entity,dataHandler)
        self.needsStretching = False

    @abstractmethod
    def matplotlibRender(self):
        pass

    @property
    def useMpld3(self):
        return mpld3Available and self.options.get("mpld3", "false") == "true"

    def canStretch(self):
        return True

    def isStretchingOn(self):
        return self.getBooleanOption('stretch', False) and not self.useMpld3

    @commonChartOptions
    def getChartOptions(self):
        options = []
        if mpld3Available:
            options.append({
                'name': 'mpld3',
                'refresh': True,
                'description': 'D3 Rendering (mpld3)',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
            })
        if self.needsStretching:
            options.append({
                'name': 'stretch',
                'description': 'Stretch image',
                'metadata':{
                    'type': 'checkbox',
                    'default': 'false'
                }
            })
        return options

    def setTicks(self, fig, ax):
        if self.handlerId=="lineChart" and self.getBooleanOption("logy",False):
            start, end = ax.get_ylim()
            ax.yaxis.set_minor_locator(ticker.MultipleLocator((end - start) / 4))
            ax.yaxis.set_minor_formatter(ticker.LogFormatter(labelOnlyBase=False))
            ax.grid(True, which='both')

        labels = [s.get_text() for s in ax.get_xticklabels()]
        totalWidth = sum(len(s) for s in labels) * 5
        if totalWidth > self.getPreferredOutputWidth():
            self.needsStretching = self.canStretch()
            if self.isStretchingOn():
                #resize image width
                fig.set_size_inches(min(totalWidth/self.getDPI(), 32768/self.getDPI()),fig.get_figheight())
            else:
                self.addMessage("Some labels are not displayed because of a lack of space. Click on Stretch image to see them all")
                #filter down the list to max 20
                max = int(len(labels)/20)
                if max < 20:
                    xl = [(i,a[:10] + ".." if len(a) > 10 else a) for i,a in enumerate(labels)]
                else:
                    xl = [(i,a) for i,a in enumerate(labels) if (i % max == 0)]
                ax.set_xticks([x[0] for x in xl])
                ax.set_xticklabels([x[1] for x in xl])
                plt.xticks(rotation=30)
        else:
            plt.xticks(rotation=0)

    def setLegend(self, fig, ax):
        if ax.get_legend() is not None:
            numLabels = len(ax.get_legend_handles_labels()[1])
            nCol = int(min(max(math.sqrt( numLabels ), 3), 6))
            nRows = int(numLabels/nCol)
            bboxPos = max(1.15, 1.0 + ((float(nRows)/2)/10.0))
            ax.legend(loc='upper center', bbox_to_anchor=(0.5, bboxPos),ncol=nCol, fancybox=True, shadow=True)

    def getNumFigures(self):
        return 1    #default, subclasses can override

    def createFigure(self):
        numFigures = self.getNumFigures()
        if numFigures <= 1:
            return plt.subplots(figsize=( int(self.getPreferredOutputWidth()/ self.getDPI()), int(self.getPreferredOutputHeight() / self.getDPI()) ))

        numCols = 1 if self.isStretchingOn() else 2 #number of columns for a multiplots, TODO make the layout configurable
        numRows = int( numFigures/numCols ) + numFigures % numCols
        imageHeight =  ((self.getPreferredOutputWidth()/numCols) * self.getHeightWidthRatio()) * numRows
        fig,ax = plt.subplots(numRows, numCols, figsize=( int(self.getPreferredOutputWidth()/self.getDPI()), int(imageHeight/self.getDPI() )))
        if numFigures%numCols != 0:
            fig.delaxes(ax.item(numFigures))
            ax = np.delete(ax,numFigures)
        return (fig,ax)

    def getSubplotHSpace(self):
        return 0.5 if self.isStretchingOn() else 0.3
        
    def doRenderChart(self):
        self.colormap = cm.jet

        # set defaults for plot
        plt.rcParams['savefig.dpi'] = 96
        plt.rcParams['font.family'] = "serif"
        plt.rcParams['font.serif'] = "cm"
        
        fig = None
        try:
            # go
            fig, ax = self.createFigure()

            if self.useMpld3:
                #TODO: rework this piece
                #keyFieldLabels = self.getKeyFieldLabels()
                #if (len(keyFieldLabels) > 0 and self.supportsKeyFieldLabels(self.handlerId) and self.supportsAggregation(self.handlerId)):
                #    plugins.connect(fig, ChartPlugin(self, keyFieldLabels))
                plugins.connect(fig, DialogPlugin(self, self.handlerId, self.dialogBody))

            #let subclass do the actual rendering
            newAx = self.matplotlibRender(fig, ax)
            if newAx is not None:
                ax = newAx

            axes = ax
            if not isinstance(axes, (list,np.ndarray)):
                axes = np.asarray([axes])

            #finalize the chart
            for i, a in np.ndenumerate(axes):
                if a.title is None or not a.title.get_visible() or a.title.get_text() == '':
                    self.setLegend(fig, a)
                self.setTicks(fig, a)
            sharex = True if len(axes) <=1 else len([a for a in axes if a._sharex is not None]) > 0
            if len(axes)>1 and not sharex:
                #adjust the height between subplots
                plt.subplots_adjust(hspace=self.getSubplotHSpace())

            #mpld3 has a bug when autolocator are used. Change to FixedLocators
            if self.useMpld3:
                self.addMessage("Warning: While great, D3 rendering is using MPLD3 library which has limitations that have not yet been fixed")
                from matplotlib import ticker
                self.debug("Converting to FixedLocator for mpld3")
                for a in axes:
                    locator = a.xaxis.get_major_locator()
                    if not isinstance(locator, ticker.FixedLocator):
                        vmin,vmax = a.xaxis.get_data_interval()
                        a.xaxis.set_major_locator(ticker.FixedLocator(locator.tick_values(vmin, vmax)))

            #Render the figure
            return self.renderFigure(fig)
        finally:
            if fig is not None:
                plt.close(fig)

    def renderFigure(self, fig):
        def genMarkup(chartFigure):
            return self.env.from_string("""
                    {0}
                    {{%for message in messages%}}
                        <div>{{{{message}}}}</div>
                    {{%endfor%}}
                """.format(chartFigure)
            ).render(messages=self.messages)
            
        if not self.useMpld3:
            import base64
            try:
                from io import BytesIO as pngIO
            except ImportError:
                from StringIO import StringIO as pngIO
            png=pngIO()
            plt.savefig(png, pad_inches=0.05, bbox_inches='tight', dpi=self.getDPI())
            try:
                return( 
                    genMarkup("""
                            <center><img style="max-width:initial !important" src="data:image/png;base64,{0}"  class="pd_save"></center>
                        """.format(base64.b64encode(png.getvalue()).decode("ascii"))
                    )
                )
            finally:
                png.close()
        else:
            mpld3.enable_notebook()
            try:
                return genMarkup(mpld3.fig_to_html(fig))
            finally:
                mpld3.disable_notebook()

    def connectElementInfo(self, element, data):
        if not hasattr(element, "get_figure") and hasattr(element,"get_children"):
            for i,child in enumerate(element.get_children()):
                plugins.connect(child.get_figure(), ElementInfoPlugin(child,data[i]))
        elif hasattr(element, "get_figure"):
            plugins.connect(element.get_figure(), ElementInfoPlugin(element, data ))

    """
    Helper to safely access i position of an ax object
    """
    def getAxItem(self, ax, pos):
        if isinstance(ax, np.ndarray):
            return ax.item(pos)
        elif pos == 0:
            return ax
        raise ValueError("Trying to access pos {} from a single ax".format(pos))
